#ifndef FSM_BITS_FSM_IMPL_HPP
#define FSM_BITS_FSM_IMPL_HPP

#include <iostream>

#include <boost/mpl/fold.hpp>
#include <boost/mpl/contains.hpp>
#include <boost/mpl/distance.hpp>
#include <boost/mpl/lambda.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/mpl/comparison.hpp>
#include <boost/mpl/arithmetic.hpp>

#include <boost/mpl/placeholders.hpp>

#include <fsm/bits/epsilon_equivalence.hpp>

#include <boost/mpl/integral_c.hpp>

#define LINEAR_SEARCH
//#define BINARY_SEARCH
//#define INDEX_SEARCH

// Az automata valos allapotai az eredeti allapotok reszhalmazaibol
// allnak ossze. Minden lepesben azt hatarozzuk meg, hogy az adott
// halmazbol adott betun keresztul melyik masik allapotokba juthatunk
// el. A check() mindig atjuttatja az aktualis allapot halmazbol a
// kovetkezo allapot halmazba az automatankat az aktualis betu
// szerint, vagy ha a szoveg vegere jutottunk, akkor azt vizsgaljuk,
// hogy az aktualis halmaznak es a veghalmaznak van-e metszete.

// Harom lehetoseg van az aktualis betuhoz tartozo allapot atmenet
// kiszamitasara, linearis kereses, binaris kereses vagy
// indexeles. Ezt a BINARY_SEARCH, LINEAR_SEARCH vagy az INDEX_SEARCH
// makroval lehet allitani. Az utobbi forditasi idoben egy tombot
// csinal, mely pointereket tartalmaz a kovetkezo elemet ellenorzo
// fuggvenyre. A tombot precompiler meta programminggal generaljuk,
// a tomb elemeit pedig template metaprogramminggal szamitjuk.

namespace fsm {

  namespace bits {

    using namespace boost;
    using namespace boost::mpl;
    using namespace boost::mpl::placeholders;

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess>
    struct Step;

    template <typename _State, typename _Char, 
              typename _Transitions, typename _StateLess>
    struct StateSetFromStateAndChar {

      typedef typename 
      fold<_Transitions, vector0<>,                       
           if_<and_<is_same<_State, Source<_2> >, is_same<_Char, Char<_2> > >,
               ordered_merge<_1, EpsilonEquivalence<Target<_2>, _Transitions>, 
                             _StateLess>, _1> >::type type;

    };
  
    template <typename _StateSet, typename _Char, 
              typename _Transitions, typename _StateLess>
    struct StateSetFromStateSetAndChar {
      
      typedef typename 
      fold<_StateSet, vector0<>, 
           ordered_merge<_1, StateSetFromStateAndChar<_2, _Char, _Transitions, 
                                                   _StateLess>, 
                         _StateLess> >::type type;
  
    };

    template <typename _State, typename _Transitions>
    struct CharSetFromState {

      typedef typename lambda<less<_1, _2> >::type CharLess;

      typedef typename 
      fold<_Transitions, vector0<>,                       
           if_<and_<is_same<_State, Source<_2> >, 
                    not_<is_same<Epsilon, Char<_2> > > >, 
               ordered_insert<_1, Char<_2>, CharLess>, _1> 
      >::type type;
      
    };

    template <typename _StateSet, typename _Transitions>
    struct CharSetFromStateSet {
      
      typedef typename lambda<less<_1, _2> >::type CharLess;

      typedef typename 
      fold<_StateSet, vector0<>, 
           ordered_merge<_1, CharSetFromState<_2, _Transitions>, CharLess>
      >::type type;
    };

    template <typename _StateSet, typename _Transitions, typename _Char>
    struct CharStepFromStateSet {
      typedef typename
      contains<typename CharSetFromStateSet<_StateSet, _Transitions>::type, _Char>::type type;
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess,
              typename _Char, bool step =
              CharStepFromStateSet<_StateSet, _Transitions, _Char>::type::value> 
    struct CharStep {
      static bool check(const char *ptr) {
        typedef typename StateSetFromStateSetAndChar<_StateSet, 
          _Char, _Transitions, _StateLess>::type TargetStateSet;
        return Step<TargetStateSet, _Stop, _Transitions, _StateLess>::check(ptr + 1);
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess,
              typename _Char>
    struct CharStep<_StateSet, _Stop, _Transitions, _StateLess, _Char, false> {
      static bool check(const char *) {
        return false;
      }
    };

    template <typename _StateSet, typename _Stop>
    struct StopStateSet {

      typedef typename
      fold<_StateSet, bool_<false>,
           or_<contains<_Stop, _2>, _1> >::type type;
    };

    template <typename _StateSet, typename _Stop>
    struct StopStep {
      static bool check() { 
        return StopStateSet<_StateSet, _Stop>::type::value; 
      }
    };


#ifdef BINARY_SEARCH

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess, 
              typename _Begin, typename _End>
    struct TransitionStepItem {
      static bool check(const char *ptr) {
        typedef typename advance<_Begin, 
          typename divides<
          typename distance<_Begin, _End>::type, int_<2> >::type >::type
          _Current;
        if (deref<_Current>::type::value > *ptr) {
          return TransitionStepItem<_StateSet, _Stop, _Transitions, _StateLess, 
            typename next<_Current>::type, _End>::check(ptr);
        } else {
          return TransitionStepItem<_StateSet, _Stop, _Transitions, _StateLess, 
            _Begin, _Current>::check(ptr);
        }
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess, 
              typename _Current>
    struct TransitionStepItem<_StateSet, _Stop, _Transitions, _StateLess, 
                          _Current, _Current> {
      static bool check(const char *ptr) {
        if (deref<_Current>::type::value == *ptr) {
          typedef typename StateSetFromStateSetAndChar<_StateSet, 
            typename deref<_Current>::type, 
            _Transitions, _StateLess>::type TargetStateSet;
          return Step<TargetStateSet, _Stop, 
            _Transitions, _StateLess>::check(ptr + 1);
        } else {
          return false;
        }
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess, 
              typename _Begin, typename _End>
    struct TransitionStep {
      static bool check(const char *ptr) {        
        return TransitionStepItem<_StateSet, _Stop, _Transitions, _StateLess,
          _Begin, typename prior<_End>::type>::check(ptr);
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess, 
              typename _End>
    struct TransitionStep<_StateSet, _Stop, _Transitions, 
                          _StateLess, _End, _End> {
      static bool check(const char *) {        
        return false;
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess>
    struct Step {
      static bool check(const char *ptr) {        
        if (*ptr != '\0') {
          typedef typename CharSetFromStateSet<_StateSet, 
            _Transitions>::type Chars;
          return TransitionStep<_StateSet, _Stop, _Transitions, _StateLess,
            typename begin<Chars>::type, typename end<Chars>::type>::check(ptr);   
        } else {
          return StopStep<_StateSet, _Stop>::check();
        }
      }
    };    

#endif

#ifdef LINEAR_SEARCH 

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess, 
              typename _Begin, typename _End>
    struct TransitionStep {
      static bool check(const char *ptr) {
        if (deref<_Begin>::type::value == *ptr) {
          typedef typename StateSetFromStateSetAndChar<_StateSet, 
            typename deref<_Begin>::type, 
            _Transitions, _StateLess>::type TargetStateSet;
          return Step<TargetStateSet, _Stop, 
            _Transitions, _StateLess>::check(ptr + 1);
        } else {
          return TransitionStep<_StateSet, _Stop, _Transitions, _StateLess, 
            typename next<_Begin>::type, _End>::check(ptr);
        }
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess, 
              typename _End>
    struct TransitionStep<_StateSet, _Stop, _Transitions, _StateLess, 
                          _End, _End> {
      static bool check(const char *) {
        return false;
      }
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess>
    struct Step {
      static bool check(const char *ptr) {        
        if (*ptr != '\0') {
          typedef typename CharSetFromStateSet<_StateSet, 
            _Transitions>::type Chars;
          return TransitionStep<_StateSet, _Stop, _Transitions, _StateLess,
            typename begin<Chars>::type, typename end<Chars>::type>::check(ptr);   
        } else {
          return StopStep<_StateSet, _Stop>::check();
        }
      }
    };    

#endif

#ifdef INDEX_SEARCH

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess,
              char _char>
    struct TransitionStep {
      static bool check(const char *ptr) {
        return CharStep<_StateSet, _Stop, _Transitions, _StateLess,
          integral_c<char, _char> >::check(ptr);
      }      
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess>
    struct TransitionStep<_StateSet, _Stop, _Transitions, _StateLess, '\0'> {
      static bool check(const char *) {
        return StopStep<_StateSet, _Stop>::check();
      }      
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess>
    struct Step {
      typedef bool (*CheckFun)(const char*);
      static const CheckFun trans[];
      
      static bool check(const char *ptr) {
        return trans[(int)*ptr - CHAR_MIN](ptr);
      }
      
    };

    template <typename _StateSet, typename _Stop, 
              typename _Transitions, typename _StateLess>
    const typename Step<_StateSet, _Stop, _Transitions, _StateLess>::CheckFun 
    Step<_StateSet, _Stop, _Transitions, _StateLess>::trans[] = {
#include <fsm/bits/index_search_iteration.hpp>
    }; 
#endif

  }
}

#endif
